#include "io/img/Img.h"

namespace ehs
{
	Array<ImgCodec> Img::codecs;

	bool Img::HasCodec(const UInt_64 hashExt)
	{
		for (UInt_64 i = 0; i < codecs.Size(); ++i)
			if (codecs[i].GetHashExt() == hashExt)
				return true;

		return false;
	}

	bool Img::HasCodec(const Str_8& ext)
	{
		return HasCodec(ext.Hash_64());
	}

	bool Img::AddCodec(ImgCodec codec)
	{
		if (HasCodec(codec.GetHashExt()))
			return false;

		codecs.Push(std::move(codec));

		return true;
	}

	const ImgCodec* Img::GetCodec(const UInt_64 hashExt)
	{
		for (UInt_64 i = 0; i < codecs.Size(); ++i)
			if (codecs[i].GetHashExt() == hashExt)
				return &codecs[i];

		return nullptr;
	}

	const ImgCodec* Img::GetCodec(const Str_8& ext)
	{
		return GetCodec(ext.Hash_64());
	}

	Img::~Img()
	{
		delete[] data;
	}

	Img::Img()
		: data(nullptr), bitDepth(0), channels(0), width(0), height(0), size(0)
	{
	}

	Img::Img(const UInt_8 bitDepth, const UInt_8 channels, const UInt_64 width, const UInt_64 height, const Byte* const data)
		: bitDepth(bitDepth), channels(channels), width(width), height(height),
		size(width * (bitDepth / 8) * channels * height), data(new Byte[size])
	{
		Util::Copy(this->data, data, size);
	}

	Img::Img(const UInt_8 bitDepth, const UInt_8 channels, const UInt_64 width, const UInt_64 height)
		 : bitDepth(bitDepth), channels(channels), width(width), height(height),
		 size(width * (bitDepth / 8) * channels * height), data(new Byte[size])
	{
	}

	Img::Img(Img&& img) noexcept
		: bitDepth(img.bitDepth), channels(img.channels), width(img.width), height(img.height), size(img.size),
		data(img.data)
	{
		img.bitDepth = 0;
		img.channels = 0;
		img.width = 0;
		img.height = 0;
		img.size = 0;
		img.data = nullptr;
	}

	Img::Img(const Img& img)
		: bitDepth(img.bitDepth), channels(img.channels), width(img.width), height(img.height), size(img.size),
		data(new Byte[img.size])
	{
		Util::Copy(data, img.data, img.size);
	}

	Img& Img::operator=(Img&& img) noexcept
	{
		if (this == &img)
			return *this;

		Release();

		bitDepth = img.bitDepth;
		channels = img.channels;
		width = img.width;
		height = img.height;
		size = img.size;
		data = img.data;

		img.bitDepth = 0;
		img.channels = 0;
		img.width = 0;
		img.height = 0;
		img.size = 0;
		img.data = nullptr;

		return *this;
	}

	Img& Img::operator=(const Img& img)
	{
		if (this == &img)
			return *this;

		Release();

		bitDepth = img.bitDepth;
		channels = img.channels;
		width = img.width;
		height = img.height;
		size = img.size;
		delete[] data;
		data = new Byte[img.size];
		Util::Copy(data, img.data, img.size);

		return *this;
	}

	Img::operator const Byte* () const
	{
		return data;
	}

	Img::operator Byte* ()
	{
		return data;
	}

	void Img::Release()
	{
		delete[] data;
		data = nullptr;
	}

	UInt_8 Img::BitDepth() const
	{
		return bitDepth;
	}

	UInt_8 Img::Channels() const
	{
		return channels;
	}

	UInt_64 Img::Width() const
	{
		return width;
	}

	UInt_64 Img::Height() const
	{
		return height;
	}

	UInt_64 Img::Size() const
	{
		return size;
	}

	void Img::SetPixel(const UInt_64 index, const Byte* const pixel)
	{
		UInt_64 rIndex = index * (bitDepth / 8) * channels;

		for (UInt_64 i = 0; i < bitDepth / 8 * channels; ++i)
			data[rIndex + i] = pixel[i];
	}

	void Img::GetPixel(const UInt_64 index, Byte* const pixel) const
	{
		UInt_64 rIndex = index * (bitDepth / 8) * channels;

		for (UInt_64 i = 0; i < bitDepth / 8 * channels; ++i)
			pixel[i] = data[rIndex + i];
	}

	void Img::SetPixel(const UInt_64 x, const UInt_64 y, const Byte* const pixel)
	{
		UInt_64 rIndex = (y * width * (bitDepth / 8) * channels) + (x * (bitDepth / 8) * channels);

		for (UInt_64 i = 0; i < bitDepth / 8 * channels; ++i)
			data[rIndex + i] = pixel[i];
	}

	void Img::GetPixel(const UInt_64 x, const UInt_64 y, Byte* const pixel) const
	{
		UInt_64 rIndex = (y * width * (bitDepth / 8) * channels) + (x * (bitDepth / 8) * channels);

		for (UInt_64 i = 0; i < bitDepth / 8 * channels; ++i)
			pixel[i] = data[rIndex + i];
	}

	void Img::Resize(const Resampling method, const UInt_64 newWidth, const UInt_64 newHeight)
	{
		switch (method)
		{
			case Resampling::NEAREST_NEIGHBOR:
			{
				NearestNeighbor(newWidth, newHeight);
				return;
			}
			default:
				return;
		}
	}

	Img Img::GetResized(const Resampling method, const UInt_64 newWidth, const UInt_64 newHeight) const
	{
		switch (method)
		{
			case Resampling::NEAREST_NEIGHBOR:
				return GetNearestNeighbor(newWidth, newHeight);
			default:
				return {};
		}
	}

	void Img::ToRGBA()
	{
		switch (channels)
		{
			case 3:
			{
				size = width * height * (bitDepth / 8) * 4;
				Byte* result = new Byte[size];
				RGB_To_RGBA(size, result);
				channels = 4;
				delete[] data;
				data = result;
				break;
			}
			case 2:
			{
				size = width * height * (bitDepth / 8) * 4;
				Byte* result = new Byte[size];
				MonoA_To_RGBA(size, result);
				channels = 4;
				delete[] data;
				data = result;
				break;
			}
			case 1:
			{
				size = width * height * (bitDepth / 8) * 4;
				Byte* result = new Byte[size];
				Mono_To_RGBA(size, result);
				channels = 4;
				delete[] data;
				data = result;
				break;
			}
			default:
			{
				return;
			}
		}
	}

	Img Img::GetAsRGBA() const
	{
		switch (channels)
		{
			case 4:
			{
				return {*this};
			}
			case 3:
			{
				Img result(bitDepth, 4, width, height);
				RGB_To_RGBA(result.Size(), result);
				return result;
			}
			case 2:
			{
				Img result(bitDepth, 4, width, height);
				MonoA_To_RGBA(result.Size(), result);
				return result;
			}
			case 1:
			{
				Img result(bitDepth, 4, width, height);
				Mono_To_RGBA(result.Size(), result);
				return result;
			}
			default:
			{
				return {};
			}
		}
	}

	void Img::ToRGB()
	{
		switch (channels)
		{
			case 4:
			{
				size = width * height * (bitDepth / 8) * 3;
				Byte* result = new Byte[size];
				RGBA_To_RGB(size, result);
				channels = 3;
				delete[] data;
				data = result;
				break;
			}
			case 2:
			{
				size = width * height * (bitDepth / 8) * 3;
				Byte* result = new Byte[size];
				MonoA_To_RGB(size, result);
				channels = 3;
				delete[] data;
				data = result;
				break;
			}
			case 1:
			{
				size = width * height * (bitDepth / 8) * 3;
				Byte* result = new Byte[size];
				Mono_To_RGB(size, result);
				channels = 3;
				delete[] data;
				data = result;
				break;
			}
			default:
			{
				return;
			}
		}
	}

	Img Img::GetAsRGB() const
	{
		switch (channels)
		{
			case 4:
			{
				Img result(bitDepth, 3, width, height);
				RGBA_To_RGB(result.Size(), result);
				return result;
			}
			case 3:
			{
				return {*this};
			}
			case 2:
			{
				Img result(bitDepth, 3, width, height);
				MonoA_To_RGB(result.Size(), result);
				return result;
			}
			case 1:
			{
				Img result(bitDepth, 3, width, height);
				Mono_To_RGB(result.Size(), result);
				return result;
			}
			default:
			{
				return {};
			}
		}
	}

	void Img::ToMonoA()
	{
		switch (channels)
		{
			case 4:
			{
				size = width * height * (bitDepth / 8) * 2;
				Byte* result = new Byte[size];
				RGBA_To_MonoA(size, result);
				channels = 2;
				delete[] data;
				data = result;
				break;
			}
			case 3:
			{
				size = width * height * (bitDepth / 8) * 2;
				Byte* result = new Byte[size];
				RGB_To_MonoA(size, result);
				channels = 2;
				delete[] data;
				data = result;
				break;
			}
			case 1:
			{
				size = width * height * (bitDepth / 8) * 2;
				Byte* result = new Byte[size];
				Mono_To_MonoA(size, result);
				channels = 2;
				delete[] data;
				data = result;
				break;
			}
			default:
			{
				return;
			}
		}
	}

	Img Img::GetAsMonoA() const
	{
		switch (channels)
		{
			case 4:
			{
				Img result(bitDepth, 2, width, height);
				RGBA_To_MonoA(result.Size(), result);
				return result;
			}
			case 3:
			{
				Img result(bitDepth, 2, width, height);
				RGB_To_MonoA(result.Size(), result);
				return result;
			}
			case 2:
			{
				return {*this};
			}
			case 1:
			{
				Img result(bitDepth, 2, width, height);
				Mono_To_MonoA(result.Size(), result);
				return result;
			}
			default:
			{
				return {};
			}
		}
	}

	void Img::ToMono()
	{
		switch (channels)
		{
			case 4:
			{
				size = width * height * (bitDepth / 8);
				Byte* result = new Byte[size];
				RGBA_To_Mono(size, result);
				channels = 1;
				delete[] data;
				data = result;
				break;
			}
			case 3:
			{
				size = width * height * (bitDepth / 8);
				Byte* result = new Byte[size];
				RGB_To_Mono(size, result);
				channels = 1;
				delete[] data;
				data = result;
				break;
			}
			case 2:
			{
				size = width * height * (bitDepth / 8);
				Byte* result = new Byte[size];
				MonoA_To_Mono(size, result);
				channels = 1;
				delete[] data;
				data = result;
				break;
			}
			default:
			{
				return;
			}
		}
	}

	Img Img::GetAsMono() const
	{
		switch (channels)
		{
			case 4:
			{
				Img result(bitDepth, 1, width, height);
				RGBA_To_Mono(result.Size(), result);
				return result;
			}
			case 3:
			{
				Img result(bitDepth, 1, width, height);
				RGB_To_Mono(result.Size(), result);
				return result;
			}
			case 2:
			{
				Img result(bitDepth, 1, width, height);
				MonoA_To_Mono(result.Size(), result);
				return result;
			}
			case 1:
			{
				return {*this};
			}
			default:
			{
				return {};
			}
		}
	}

	void Img::To32()
	{
		switch (bitDepth)
		{
			case 24:
			{
				size = width * height * 4 * channels;
				Byte* result = new Byte[size];
				BD24_to_BD32(size, result);
				bitDepth = 32;
				delete[] data;
				data = result;
				break;
			}
			case 16:
			{
				size = width * height * 4 * channels;
				Byte* result = new Byte[size];
				BD16_to_BD32(size, result);
				bitDepth = 32;
				delete[] data;
				data = result;
				break;
			}
			case 8:
			{
				size = width * height * 4 * channels;
				Byte* result = new Byte[size];
				BD8_to_BD32(size, result);
				bitDepth = 32;
				delete[] data;
				data = result;
				break;
			}
			default:
			{
				return;
			}
		}
	}

	Img Img::GetAs32() const
	{
		switch (bitDepth)
		{
			case 32:
			{
				return {*this};
			}
			case 24:
			{
				Img result(32, channels, width, height);
				BD24_to_BD32(result.Size(), result);
				return result;
			}
			case 16:
			{
				Img result(32, channels, width, height);
				BD16_to_BD32(result.Size(), result);
				return result;
			}
			case 8:
			{
				Img result(32, channels, width, height);
				BD8_to_BD32(result.Size(), result);
				return result;
			}
			default:
			{
				return {};
			}
		}
	}

	void Img::To24()
	{
		switch (bitDepth)
		{
			case 32:
			{
				size = width * height * 3 * channels;
				Byte* result = new Byte[size];
				BD32_to_BD24(size, result);
				bitDepth = 24;
				delete[] data;
				data = result;
				break;
			}
			case 16:
			{
				size = width * height * 3 * channels;
				Byte* result = new Byte[size];
				BD16_to_BD24(size, result);
				bitDepth = 24;
				delete[] data;
				data = result;
				break;
			}
			case 8:
			{
				size = width * height * 3 * channels;
				Byte* result = new Byte[size];
				BD8_to_BD24(size, result);
				bitDepth = 24;
				delete[] data;
				data = result;
				break;
			}
			default:
			{
				return;
			}
		}
	}

	Img Img::GetAs24() const
	{
		switch (bitDepth)
		{
			case 32:
			{
				Img result(24, channels, width, height);
				BD32_to_BD24(result.Size(), result);
				return result;
			}
			case 24:
			{
				return {*this};
			}
			case 16:
			{
				Img result(24, channels, width, height);
				BD16_to_BD24(result.Size(), result);
				return result;
			}
			case 8:
			{
				Img result(24, channels, width, height);
				BD8_to_BD24(result.Size(), result);
				return result;
			}
			default:
			{
				return {};
			}
		}
	}

	void Img::To16()
	{
		switch (bitDepth)
		{
			case 32:
			{
				size = width * height * 2 * channels;
				Byte* result = new Byte[size];
				BD32_to_BD16(size, result);
				bitDepth = 16;
				delete[] data;
				data = result;
				break;
			}
			case 24:
			{
				size = width * height * 2 * channels;
				Byte* result = new Byte[size];
				BD24_to_BD16(size, result);
				bitDepth = 16;
				delete[] data;
				data = result;
				break;
			}
			case 8:
			{
				size = width * height * 2 * channels;
				Byte* result = new Byte[size];
				BD8_to_BD16(size, result);
				bitDepth = 16;
				delete[] data;
				data = result;
				break;
			}
			default:
			{
				return;
			}
		}
	}

	Img Img::GetAs16() const
	{
		switch (bitDepth)
		{
			case 32:
			{
				Img result(16, channels, width, height);
				BD32_to_BD16(result.Size(), result);
				return result;
			}
			case 24:
			{
				Img result(16, channels, width, height);
				BD24_to_BD16(result.Size(), result);
				return result;
			}
			case 16:
			{
				return {*this};
			}
			case 8:
			{
				Img result(16, channels, width, height);
				BD8_to_BD16(result.Size(), result);
				return result;
			}
			default:
			{
				return {};
			}
		}
	}

	void Img::To8()
	{
		switch (bitDepth)
		{
			case 32:
			{
				size = width * height * channels;
				Byte* result = new Byte[size];
				BD32_to_BD8(size, result);
				bitDepth = 8;
				delete[] data;
				data = result;
				break;
			}
			case 24:
			{
				size = width * height * channels;
				Byte* result = new Byte[size];
				BD24_to_BD8(size, result);
				bitDepth = 8;
				delete[] data;
				data = result;
				break;
			}
			case 16:
			{
				size = width * height * channels;
				Byte* result = new Byte[size];
				BD16_to_BD8(size, result);
				bitDepth = 8;
				delete[] data;
				data = result;
				break;
			}
			default:
			{
				return;
			}
		}
	}

	Img Img::GetAs8() const
	{
		switch (bitDepth)
		{
			case 32:
			{
				Img result(8, channels, width, height);
				BD32_to_BD8(result.Size(), result);
				return result;
			}
			case 24:
			{
				Img result(8, channels, width, height);
				BD24_to_BD8(result.Size(), result);
				return result;
			}
			case 16:
			{
				Img result(8, channels, width, height);
				BD16_to_BD8(result.Size(), result);
				return result;
			}
			case 8:
			{
				return {*this};
			}
			default:
			{
				return {};
			}
		}
	}

	bool Img::IsValid() const
	{
		return width && height;
	}

	bool Img::ToFile(const Str_8& filePath) const
	{
		Str_8 ext = File::ProcessExt_8(filePath);

		const ImgCodec* codec = GetCodec(ext);
		if (!codec)
		{
			EHS_LOG_INT("Error", 0, "Codec not found for file extension, \"" + ext + "\".");
			return false;
		}

		Serializer<UInt_64> result;
		if (!codec->Encode(result, this))
			return false;

		File file(filePath, Mode::WRITE, Disposition::CREATE_PERSISTENT);
		file.WriteSerializer_64(result);

		return true;
	}

	Img Img::FromFile(const Str_8& filePath)
	{
		File file(filePath, Mode::READ, Disposition::OPEN);
		Str_8 ext = file.GetExtension();

		Img result;

		const ImgCodec* codec = GetCodec(ext);
		if (!codec)
		{
			EHS_LOG_INT("Error", 0, "Codec not found for file extension, \"" + ext + "\".");
			return result;
		}

		Serializer<UInt_64> data = file.ReadSerializer_64(codec->GetEndianness(), file.Size());

		file.Release();

		if (!codec->Decode(data, &result))
			return {};

		return result;
	}

	Img* Img::FromFile_Heap(const Str_8& filePath)
	{
		File file(filePath, Mode::READ, Disposition::OPEN);
		Str_8 ext = file.GetExtension();

		Img* result = nullptr;

		const ImgCodec* codec = GetCodec(ext);
		if (!codec)
		{
			EHS_LOG_INT("Error", 0, "Codec not found for file extension, \"" + ext + "\".");
			return result;
		}

		result = new Img();

		Serializer<UInt_64> data = file.ReadSerializer_64(codec->GetEndianness(), file.Size());

		file.Release();

		if (!codec->Decode(data, result))
		{
			delete result;
			return nullptr;
		}

		return result;
	}

	Img Img::FromData(const Str_8& ext, Serializer<UInt_64>& data)
	{
		Img result;

		const ImgCodec* codec = GetCodec(ext);
		if (!codec)
		{
			EHS_LOG_INT("Error", 0, "Codec not found for file extension, \"" + ext + "\".");
			return result;
		}

		if (!codec->Decode(data, &result))
			return {};

		return result;
	}

	Img Img::GetNearestNeighbor(const UInt_64 newWidth, const UInt_64 newHeight) const
	{
		UInt_8 bytes = bitDepth / 8;

		Img result(bitDepth, channels, newWidth, newHeight);

		double xRatio = (double)width / (double)newWidth;
		double yRatio = (double)height / (double)newWidth;
		double px;
		double py;

		for(UInt_64 y = 0; y < newHeight; ++y)
		{
			for (UInt_64 x = 0; x < newWidth; ++x)
			{
				px = Math::Floor((double)x * xRatio);
				py = Math::Floor((double)y * yRatio);

				for (UInt_64 b = 0; b < bytes * channels; ++b)
					result[y * newWidth * bytes * channels + x * bytes * channels + b] = data[(UInt_64)py * width * bytes * channels + (UInt_64)px * bytes * channels + b];
			}
		}

		return result;
	}

	void Img::NearestNeighbor(const UInt_64 newWidth, const UInt_64 newHeight)
	{
		UInt_8 bytes = bitDepth / 8;

		Byte* result = new Byte[newWidth * bytes * channels * newHeight];

		double xRatio = (double)width / (double)newWidth;
		double yRatio = (double)height / (double)newWidth;
		double px;
		double py;

		for(UInt_64 y = 0; y < newHeight; ++y)
		{
			for (UInt_64 x = 0; x < newWidth; ++x)
			{
				px = Math::Floor((double)x * xRatio);
				py = Math::Floor((double)y * yRatio);

				for (UInt_64 b = 0; b < bytes * channels; ++b)
					result[y * newWidth * bytes * channels + x * bytes * channels + b] = data[(UInt_64)py * width * bytes * channels + (UInt_64)px * bytes * channels + b];
			}
		}

		delete[] data;
		data = result;
		width = newWidth;
		height = newHeight;
	}

	void Img::RGB_To_RGBA(const UInt_64 newSize, Byte* buffer) const
	{
		UInt_8 bytes = bitDepth / 8;

		for (UInt_64 i = 0, n = 0; i < newSize; i += bytes * 4, n += bytes * channels)
		{
			for (UInt_64 b = 0; b < bytes * 3; ++b)
				buffer[i + b] = data[n + b];

			if (bytes == 1)
				buffer[i + bytes * 3] = EHS_UINT_8_MAX;
			else if (bytes == 2)
				*(UInt_16*)&buffer[i + bytes * 3] = EHS_UINT_16_MAX;
			else if (bytes == 3)
			{
				UInt_32 value = EHS_UINT_24_MAX;
				for (UInt_64 b = 0; b < bytes; ++b)
					buffer[i + bytes * 3 + b] = ((Byte*)&value)[b];
			}
			else if (bytes == 4)
				*(UInt_32*)&buffer[i + bytes * 3] = EHS_UINT_32_MAX;
		}
	}

	void Img::MonoA_To_RGBA(const UInt_64 newSize, Byte* buffer) const
	{
		UInt_8 bytes = bitDepth / 8;

		for (UInt_64 i = 0, n = 0; i < newSize; i += bytes * 4, n += bytes * channels)
		{
			for (UInt_64 b = 0; b < bytes * 4; ++b)
				buffer[i + b] = data[n + b % bytes];

			if (bytes == 1)
				buffer[i + bytes * 3] = data[n + bytes];
			else if (bytes == 2)
				*(UInt_16*) &buffer[i + bytes * 3] = *(UInt_16*) &data[n + bytes];
			else if (bytes == 3)
			{
				for (UInt_64 b = 0; b < bytes; ++b)
					buffer[i + bytes * 3 + b] = data[n + bytes + b];
			}
			else if (bytes == 4)
				*(UInt_32*) &buffer[i + bytes * 3] = *(UInt_32*) &data[n + bytes];
		}
	}

	void Img::Mono_To_RGBA(const UInt_64 newSize, Byte* buffer) const
	{
		UInt_8 bytes = bitDepth / 8;

		for (UInt_64 i = 0, n = 0; i < newSize; i += bytes * 4, n += bytes * channels)
		{
			for (UInt_64 b = 0; b < bytes * 4; ++b)
				buffer[i + b] = data[n + b % bytes];

			if (bytes == 1)
				buffer[i + bytes * 3] = EHS_UINT_8_MAX;
			else if (bytes == 2)
				*(UInt_16*)&buffer[i + bytes * 3] = EHS_UINT_16_MAX;
			else if (bytes == 3)
			{
				UInt_32 value = EHS_UINT_24_MAX;
				for (UInt_64 b = 0; b < bytes; ++b)
					buffer[i + bytes * 3 + b] = ((Byte*)&value)[b];
			}
			else if (bytes == 4)
				*(UInt_32*) &buffer[i + bytes * 3] = EHS_UINT_32_MAX;
		}
	}

	void Img::RGBA_To_RGB(const UInt_64 newSize, Byte* buffer) const
	{
		UInt_8 bytes = bitDepth / 8;

		for (UInt_64 i = 0, n = 0; i < newSize; i += bytes * 3, n += bytes * channels)
			for (UInt_64 b = 0; b < bytes * 3; ++b)
				buffer[i + b] = data[n + b];
	}

	void Img::MonoA_To_RGB(const UInt_64 newSize, Byte* buffer) const
	{
		UInt_8 bytes = bitDepth / 8;

		for (UInt_64 i = 0, n = 0; i < newSize; i += bytes * 3, n += bytes * channels)
			for (UInt_64 b = 0; b < bytes * 3; ++b)
				buffer[i + b] = data[n + b % bytes];
	}

	void Img::Mono_To_RGB(const UInt_64 newSize, Byte* buffer) const
	{
		UInt_8 bytes = bitDepth / 8;

		for (UInt_64 i = 0, n = 0; i < newSize; i += bytes * 3, n += bytes * channels)
			for (UInt_64 b = 0; b < bytes * 3; ++b)
				buffer[i + b] = data[n + b % bytes];
	}

	void Img::RGBA_To_MonoA(const UInt_64 newSize, Byte* buffer) const
	{
		UInt_8 bytes = bitDepth / 8;

		for (UInt_64 i = 0, n = 0; i < newSize; i += bytes * 2, n += bytes * channels)
		{
			if (bytes == 1)
			{
				UInt_16 average = 0;

				for (UInt_64 b = 0; b < 3; ++b)
					average += data[n + b];

				buffer[i] = average / 4;
				buffer[i + bytes] = data[n + 3];
			}
			else if (bytes == 2)
			{
				UInt_32 average = 0;

				for (UInt_64 b = 0; b < bytes * 3; b += bytes)
					average += *(UInt_16*)&data[n + b];

				*(UInt_16*)&buffer[i] = average / 4;
				*(UInt_16*)&buffer[i + bytes] = *(UInt_16*)&data[n + bytes * 3];
			}
			else if (bytes == 3)
			{
				UInt_64 average = 0;

				for (UInt_64 b = 0; b < bytes * 3; b += bytes)
				{
					UInt_32 num = 0;
					num |= data[n + b];
					num |= data[n + b + 1] << 8;
					num |= data[n + b + 2] << 16;

					average += num;
				}

				average /= 4;

				buffer[i] = ((Byte*)&average)[0];
				buffer[i + 1] = ((Byte*)&average)[1];
				buffer[i + 2] = ((Byte*)&average)[2];

				buffer[i + bytes] = data[n + bytes * 3];
				buffer[i + bytes + 1] = data[n + bytes * 3 + 1];
				buffer[i + bytes + 2] = data[n + bytes * 3 + 2];
			}
			else if (bytes == 4)
			{
				UInt_64 average = 0;

				for (UInt_64 b = 0; b < bytes * 3; b += bytes)
					average += *(UInt_32*)&data[n + b];

				*(UInt_32*)&buffer[i] = average / 4;
				*(UInt_32*)&buffer[i + bytes] = *(UInt_32*)&data[n + bytes * 3];
			}
		}
	}

	void Img::RGB_To_MonoA(const UInt_64 newSize, Byte* buffer) const
	{
		UInt_8 bytes = bitDepth / 8;

		for (UInt_64 i = 0, n = 0; i < newSize; i += bytes * 2, n += bytes * channels)
		{
			if (bytes == 1)
			{
				UInt_16 average = 0;

				for (UInt_64 b = 0; b < 3; ++b)
					average += data[n + b];

				buffer[i] = average / 4;
				buffer[i + bytes] = EHS_UINT_8_MAX;
			}
			else if (bytes == 2)
			{
				UInt_32 average = 0;

				for (UInt_64 b = 0; b < bytes * 3; b += bytes)
					average += *(UInt_16*)&data[n + b];

				*(UInt_16*)&buffer[i] = average / 4;
				*(UInt_16*)&buffer[i + bytes] = EHS_UINT_16_MAX;
			}
			else if (bytes == 3)
			{
				UInt_64 average = 0;

				for (UInt_64 b = 0; b < bytes * 3; b += bytes)
				{
					UInt_32 num = 0;
					num |= data[n + b];
					num |= data[n + b + 1] << 8;
					num |= data[n + b + 2] << 16;

					average += num;
				}

				average /= 4;

				buffer[i] = ((Byte*)&average)[0];
				buffer[i + 1] = ((Byte*)&average)[1];
				buffer[i + 2] = ((Byte*)&average)[2];

				UInt_32 tmp = EHS_UINT_24_MAX;

				buffer[i + bytes] = ((Byte*)&tmp)[0];
				buffer[i + bytes + 1] = ((Byte*)&tmp)[1];
				buffer[i + bytes + 2] = ((Byte*)&tmp)[2];
			}
			else if (bytes == 4)
			{
				UInt_64 average = 0;

				for (UInt_64 b = 0; b < bytes * 3; b += bytes)
					average += *(UInt_32*)&data[n + b];

				*(UInt_32*)&buffer[i] = average / 4;
				*(UInt_32*)&buffer[i + bytes] = EHS_UINT_32_MAX;
			}
		}
	}

	void Img::Mono_To_MonoA(const UInt_64 newSize, Byte* buffer) const
	{
		UInt_8 bytes = bitDepth / 8;

		for (UInt_64 i = 0, n = 0; i < newSize; i += bytes * 2, n += bytes * channels)
		{
			if (bytes == 1)
			{
				buffer[i] = data[n];
				buffer[i + bytes] = EHS_UINT_8_MAX;
			}
			else if (bytes == 2)
			{
				*(UInt_16*)&buffer[i] = *(UInt_16*)&data[n];
				*(UInt_16*)&buffer[i + bytes] = EHS_UINT_16_MAX;
			}
			else if (bytes == 3)
			{
				buffer[i] = data[n];
				buffer[i + 1] = data[n + 1];
				buffer[i + 2] = data[n + 2];

				UInt_32 tmp = EHS_UINT_24_MAX;

				buffer[i + bytes] = ((Byte*)&tmp)[0];
				buffer[i + bytes + 1] = ((Byte*)&tmp)[1];
				buffer[i + bytes + 2] = ((Byte*)&tmp)[2];
			}
			else if (bytes == 4)
			{
				*(UInt_32*)&buffer[i] = *(UInt_32*)&data[n];
				*(UInt_32*)&buffer[i + bytes] = EHS_UINT_32_MAX;
			}
		}
	}

	void Img::RGBA_To_Mono(const UInt_64 newSize, Byte* buffer) const
	{
		UInt_8 bytes = bitDepth / 8;

		for (UInt_64 i = 0, n = 0; i < newSize; i += bytes, n += bytes * channels)
		{
			if (bytes == 1)
			{
				UInt_16 average = 0;

				for (UInt_64 b = 0; b < 3; ++b)
					average += data[n + b];

				buffer[i] = average / 4;
			}
			else if (bytes == 2)
			{
				UInt_32 average = 0;

				for (UInt_64 b = 0; b < bytes * 3; b += bytes)
					average += *(UInt_16*)&data[n + b];

				*(UInt_16*)&buffer[i] = average / 4;
			}
			else if (bytes == 3)
			{
				UInt_64 average = 0;

				for (UInt_64 b = 0; b < bytes * 3; b += bytes)
				{
					UInt_32 num = 0;
					num |= data[n + b];
					num |= data[n + b + 1] << 8;
					num |= data[n + b + 2] << 16;

					average += num;
				}

				average /= 4;

				buffer[i] = ((Byte*)&average)[0];
				buffer[i + 1] = ((Byte*)&average)[1];
				buffer[i + 2] = ((Byte*)&average)[2];
			}
			else if (bytes == 4)
			{
				UInt_64 average = 0;

				for (UInt_64 b = 0; b < bytes * 3; b += bytes)
					average += *(UInt_32*)&data[n + b];

				*(UInt_32*)&buffer[i] = average / 4;
			}
		}
	}

	void Img::RGB_To_Mono(const UInt_64 newSize, Byte* buffer) const
	{
		UInt_8 bytes = bitDepth / 8;

		for (UInt_64 i = 0, n = 0; i < newSize; i += bytes, n += bytes * channels)
		{
			if (bytes == 1)
			{
				UInt_16 average = 0;

				for (UInt_64 b = 0; b < 3; ++b)
					average += data[n + b];

				buffer[i] = average / 4;
			}
			else if (bytes == 2)
			{
				UInt_32 average = 0;

				for (UInt_64 b = 0; b < bytes * 3; b += bytes)
					average += *(UInt_16*)&data[n + b];

				*(UInt_16*)&buffer[i] = average / 4;
			}
			else if (bytes == 3)
			{
				UInt_64 average = 0;

				for (UInt_64 b = 0; b < bytes * 3; b += bytes)
				{
					UInt_32 num = 0;
					num |= data[n + b];
					num |= data[n + b + 1] << 8;
					num |= data[n + b + 2] << 16;

					average += num;
				}

				average /= 4;

				buffer[i] = ((Byte*)&average)[0];
				buffer[i + 1] = ((Byte*)&average)[1];
				buffer[i + 2] = ((Byte*)&average)[2];
			}
			else if (bytes == 4)
			{
				UInt_64 average = 0;

				for (UInt_64 b = 0; b < bytes * 3; b += bytes)
					average += *(UInt_32*)&data[n + b];

				*(UInt_32*)&buffer[i] = average / 4;
			}
		}
	}

	void Img::MonoA_To_Mono(const UInt_64 newSize, Byte* buffer) const
	{
		UInt_8 bytes = bitDepth / 8;

		for (UInt_64 i = 0, n = 0; i < newSize; i += bytes, n += bytes * channels)
		{
			if (bytes == 1)
			{
				buffer[i] = data[n];
			}
			else if (bytes == 2)
			{
				*(UInt_16*)&buffer[i] = *(UInt_16*)&data[n];
			}
			else if (bytes == 3)
			{
				buffer[i] = data[n];
				buffer[i + 1] = data[n + 1];
				buffer[i + 2] = data[n + 2];
			}
			else if (bytes == 4)
			{
				*(UInt_32*)&buffer[i] = *(UInt_32*)&data[n];
			}
		}
	}

	void Img::BD24_to_BD32(const UInt_64 newSize, Byte* buffer) const
	{
		for (UInt_64 i = 0, n = 0; i < newSize; i += 4, n += 3)
		{
			UInt_32 inValue = 0;
			((Byte*)&inValue)[0] = data[n];
			((Byte*)&inValue)[1] = data[n + 1];
			((Byte*)&inValue)[2] = data[n + 2];

			*(UInt_32*)&buffer[i] = (UInt_32)((float)inValue / (float)EHS_UINT_24_MAX * (float)EHS_UINT_32_MAX);
		}
	}

	void Img::BD16_to_BD32(const UInt_64 newSize, Byte* buffer) const
	{
		for (UInt_64 i = 0, n = 0; i < newSize; i += 4, n += 2)
			*(UInt_32*)&buffer[i] = (UInt_32)((float)*(UInt_16*)&data[n] / (float)EHS_UINT_16_MAX * (float)EHS_UINT_32_MAX);
	}

	void Img::BD8_to_BD32(const UInt_64 newSize, Byte* buffer) const
	{
		for (UInt_64 i = 0, n = 0; i < newSize; i += 4, ++n)
			*(UInt_32*)&buffer[i] = (UInt_32)((float)data[n] / (float)EHS_UINT_8_MAX * (float)EHS_UINT_32_MAX);
	}

	void Img::BD32_to_BD24(const UInt_64 newSize, Byte* buffer) const
	{
		for (UInt_64 i = 0, n = 0; i < newSize; i += 3, n += 4)
		{
			UInt_32 rValue = (UInt_32)((float)*(UInt_32*)&data[n] / (float)EHS_UINT_32_MAX * (float)EHS_UINT_24_MAX);

			buffer[i] = ((Byte*)&rValue)[0];
			buffer[i + 1] = ((Byte*)&rValue)[1];
			buffer[i + 2] = ((Byte*)&rValue)[2];
		}
	}

	void Img::BD16_to_BD24(const UInt_64 newSize, Byte* buffer) const
	{
		for (UInt_64 i = 0, n = 0; i < newSize; i += 3, n += 2)
		{
			UInt_32 rValue = (UInt_32)((float)*(UInt_16*)&data[n] / (float)EHS_UINT_16_MAX * (float)EHS_UINT_24_MAX);

			buffer[i] = ((Byte*)&rValue)[0];
			buffer[i + 1] = ((Byte*)&rValue)[1];
			buffer[i + 2] = ((Byte*)&rValue)[2];
		}
	}

	void Img::BD8_to_BD24(const UInt_64 newSize, Byte* buffer) const
	{
		for (UInt_64 i = 0, n = 0; i < newSize; i += 3, ++n)
		{
			UInt_32 rValue = (UInt_32)((float)data[n] / (float)EHS_UINT_8_MAX * (float)EHS_UINT_24_MAX);

			buffer[i] = ((Byte*)&rValue)[0];
			buffer[i + 1] = ((Byte*)&rValue)[1];
			buffer[i + 2] = ((Byte*)&rValue)[2];
		}
	}

	void Img::BD32_to_BD16(const UInt_64 newSize, Byte* buffer) const
	{
		for (UInt_64 i = 0, n = 0; i < newSize; i += 2, n += 4)
			*(UInt_16*)&buffer[i] = (UInt_16)((float)*(UInt_32*)&data[n] / (float)EHS_UINT_32_MAX * (float)EHS_UINT_16_MAX);
	}

	void Img::BD24_to_BD16(const UInt_64 newSize, Byte* buffer) const
	{
		for (UInt_64 i = 0, n = 0; i < newSize; i += 2, n += 3)
		{
			UInt_32 inValue = 0;
			((Byte*)&inValue)[0] = data[n];
			((Byte*)&inValue)[1] = data[n + 1];
			((Byte*)&inValue)[2] = data[n + 2];

			*(UInt_16*)&buffer[i] = (UInt_16)((float)inValue / (float)EHS_UINT_24_MAX * (float)EHS_UINT_16_MAX);
		}
	}

	void Img::BD8_to_BD16(const UInt_64 newSize, Byte* buffer) const
	{
		for (UInt_64 i = 0, n = 0; i < newSize; i += 2, ++n)
			*(UInt_16*)&buffer[i] = (UInt_16)((float)data[n] / (float)EHS_UINT_8_MAX * (float)EHS_UINT_16_MAX);
	}

	void Img::BD32_to_BD8(const UInt_64 newSize, Byte* buffer) const
	{
		for (UInt_64 i = 0, n = 0; i < newSize; ++i, n += 4)
			buffer[i] = (Byte)((float)*(UInt_32*)&data[n] / (float)EHS_UINT_32_MAX * (float)EHS_UINT_8_MAX);
	}

	void Img::BD24_to_BD8(const UInt_64 newSize, Byte* buffer) const
	{
		for (UInt_64 i = 0, n = 0; i < newSize; ++i, n += 3)
		{
			UInt_32 inValue = 0;
			((Byte*)&inValue)[0] = data[n];
			((Byte*)&inValue)[1] = data[n + 1];
			((Byte*)&inValue)[2] = data[n + 2];

			buffer[i] = (Byte)((float)inValue / (float)EHS_UINT_24_MAX * (float)EHS_UINT_8_MAX);
		}
	}

	void Img::BD16_to_BD8(const UInt_64 newSize, Byte* buffer) const
	{
		for (UInt_64 i = 0, n = 0; i < newSize; ++i, n += 2)
			buffer[i] = (Byte)((float)*(UInt_16*)&data[n] / (float)EHS_UINT_16_MAX * (float)EHS_UINT_8_MAX);
	}
}